/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ForecastFox.
 *
 * The Initial Developer of the Original Code is
 * Jon Stritar <jstritar@MIT.EDU>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Jon Stritar <jstritar@MIT.EDU>
 * Richard Klein <richwklein@mchsi.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// constants
const FF_TIME_NIGHT = 2;   // time that the night switches to day 
const FF_TIME_DAY   = 15;  // time the day switches to night

//global forecastfox object
var gForecastfox = null;

function ffLoad()
{
  gForecastfox = new Forecastfox();
  gForecastfox.init();
}

function ffUnload()
{
  gForecastfox.destroy();
  gForecastfox = null;
}

function ffResize(event)
{
  gForecastfox.onResize(event);
}

///////////////////////
// Forecastfox Object
function Forecastfox()
{
  return this;
}

Forecastfox.prototype =
{
  mUpdate: null,
  mDomain: "forecastfox.",
  mBranch: null,
  mProfileLoading: false,
  mLastEvent: null,
  
  init: function()
  {
    //get our XPCOM update service
    this.mUpdate = Components.classes["@mozdev.org/forecastfox/update;1"]
                             .getService(Components.interfaces.nsIForecastfoxUpdate);
    this.mProfile = Components.classes["@mozdev.org/forecastfox/profiles;1"]
              .getService(Components.interfaces.nsIForecastfoxProfiles);
    //set deck
    var deck = document.getElementById("ff-deck");
    deck.setAttribute("selectedIndex", "2");

    //setup preference service
    var prefService = "@mozilla.org/preferences-service;1";
    prefService = Components.classes[prefService].getService(Components.interfaces.nsIPrefService);
    this.mBranch = prefService.getBranch(this.mDomain);

    //add preference observer
    var pbi = this.mBranch.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    pbi.addObserver("", this, false);

    //add observer to nsForecastfox:update
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.addObserver(this, "forecastfox:update:start", false);
    os.addObserver(this, "forecastfox:update:error", false);
    os.addObserver(this, "forecastfox:update:end:current", false);
    os.addObserver(this, "forecastfox:update:end:forecast", false);
    os.addObserver(this, "forecastfox:update:end:links", false);
    os.addObserver(this, "forecastfox:profile:start", false);
    os.addObserver(this, "forecastfox:profile:end", false);
    os.addObserver(this, "forecastfox:profile:error", false);
    
    //add resize listener
    window.addEventListener("resize", ffResize, false);
    
    //get the xul elements that are used most often in here
    this._deck     = document.getElementById("ff-deck");
    this._strings  = document.getElementById("ff-bundle");
    this._current  = document.getElementById("ff-current-panel");
    this._spacer   = document.getElementById("ff-spacer");
    this._today    = document.getElementById("ff-today-panel");
    this._forecast = document.getElementById("ff-forecast-panel");

    //move it to the correct place
    setTimeout("gForecastfox.moveBox();", 5);

    //setup display - give window a chance to display before loading
    if (this.mUpdate.firstRun)
      setTimeout("gForecastfox.mUpdate.load(false);", 10);
    else
      this.updateAll();
  },

  destroy: function()
  {
    try {
      //remove preference observer
      var pbi = this.mBranch.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
      pbi.removeObserver("", this);
    } catch(e) {}

    try {
      //remove observers
      var os = Components.classes["@mozilla.org/observer-service;1"]
                         .getService(Components.interfaces.nsIObserverService);
      os.removeObserver(this, "forecastfox:update:start");
      os.removeObserver(this, "forecastfox:update:error");
      os.removeObserver(this, "forecastfox:update:end:current", false);
      os.removeObserver(this, "forecastfox:update:end:forecast", false);
      os.removeObserver(this, "forecastfox:update:end:links", false);
      os.removeObserver(this, "forecastfox:profile:start", false);
      os.removeObserver(this, "forecastfox:profile:end", false);
      os.removeObserver(this, "forecastfox:profile:error", false);
      os = null;
    } catch(e) {}

    //destroy variables
    this.mUpdate = null;
    this.mDomain = null;
    this.mBranch = null;
    this.mProfileLoading = null;
    this.mLastEvent = null;
  },

  observe: function(aSubject, aTopic, aData)
  {
    switch (aTopic) {
      case "forecastfox:update:start":
        this.updateStart();
        break;
      case "forecastfox:update:error":
        this.updateError(aData);
        break;
      case "forecastfox:update:end:current":
        this.updateCurrent();
        break;
      case "forecastfox:update:end:forecast":
        this.updateToday();
        this.updateForecast();
        break;
      case "forecastfox:update:end:links":
        this.updateLinks();
        break;
      case "forecastfox:profile:start":
        this.mProfileLoading = true;
        break;
      case "forecastfox:profile:end":
        this.mProfileLoading = false;
        this.moveBox();
        break;
      case "forecastfox:profile:error":
        this.mProfileLoading = false;
        break;
      case "nsPref:changed":
        switch (aData) {
          case "cc.panel.display":
          case "cc.panel.label":
          case "cc.tooltip.enabled":
          case "cc.tooltip.display":
          case "cc.tooltip.label":
            if (!this.mProfileLoading)
              this.updateCurrent();
            break;
          case "dayt.panel.display":
          case "dayt.panel.label":
          case "dayt.tooltip.enabled":
          case "dayt.tooltip.display":
          case "dayt.tooltip.label":
          case "dayt.panel.mode":
          case "dayt.panel.switch":
          case "dayf.panel.display":
          case "dayf.panel.label":
          case "dayf.tooltip.enabled":
          case "dayf.tooltip.display":
          case "dayf.tooltip.label":
          case "dayf.panel.mode":
          case "dayf.panel.days":
            if (!this.mProfileLoading) {
              this.updateForecast();
              this.updateToday();
            }
            break;
          case "cc.panel.enabled":
          case "dayt.panel.enabled":
          case "dayf.panel.enabled":
            if (!this.mProfileLoading)
              this.updateAll();
            break;
          case "general.bar":
          case "general.position":
            if (!this.mProfileLoading)
              this.moveBox();
            break;
        }
    }
  },

  getUTFPref: function(pref)
  {
    return this.mBranch.getComplexValue(pref,Components.interfaces.nsISupportsString).data;
  },
  
  setUTFPref: function(pref, value)
  {
    var supportsString = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
    supportsString.data = value;
    this.mBranch.setComplexValue(pref,Components.interfaces.nsISupportsString, supportsString);
  },
  
  moveBox: function()
  {
    //check if box is in the right location
    var toolbar   = document.getElementById(this.mBranch.getCharPref("general.bar"));
    var position  = this.mBranch.getIntPref("general.position");
    
    //if toolbar doesn't exist move to default location
    if (!toolbar) {
      this.mBranch.setCharPref("general.bar", "status-bar");
      this.mBranch.setIntPref("general.position", -1);
      return;
    }
    
    // bail if it is in the right place
    if (this.indexOf(toolbar,this._deck) == position)
      return;

    //remove us from parent
    this._deck.parentNode.removeChild(this._deck);

    this.insertAtIndex(toolbar, this._deck, position);
  },
  
  onResize: function(event)
  {
    // Collapses the deck if the toolbar is wider than the window and the window is not maximized.
    this._deck.removeAttribute("collapsed");
    
    var toolbar = this._deck.parentNode;
    var twidth  = 0;
    
    for (var x = 0; x < toolbar.childNodes.length; x++)
      twidth += toolbar.childNodes[x].boxObject.width;
    
    if (twidth > window.innerWidth && window.windowState != Components.interfaces.nsIDOMChromeWindow.STATE_MAXIMIZED)
      this._deck.setAttribute("collapsed", true);
  },
  
  indexOf: function(parent, child)
  {
    // -1 if it does not exist
    var children = parent.childNodes;
    for (var x = 0; x < children.length; x++)
      if (children[x] == child)
        return (x == children.length - 1) ? -1 : x;
   
    return "none";
  },
  
  insertAtIndex: function(parent, child, index)
  {
    var children = parent.childNodes;
    if ((children.length == 0) || (index >= children.length) || (index < 0))
      parent.appendChild(child);
    else
      parent.insertBefore(child, children[index]);
  },

  removeError: function(item)
  {
    switch (item) {
      case "forecast":
        item = document.getElementById("ff-forecast-error");
        break;
      case "current":
        item = document.getElementById("ff-current-error");
        break;
      case "today":
      default:
        return;
    }
    
    item.removeAttribute("value");
    item.removeAttribute("tooltiptext");
    item.setAttribute("collapsed", true);
  },
  
  removeProgress: function()
  {
    var wprogress = document.getElementById("ff-progress");
    
    //remove progress
    wprogress.setAttribute("mode", "determined");
    wprogress.setAttribute("value", "0");
    wprogress.removeAttribute("tooltiptext");
    wprogress.setAttribute("collapsed", "true");
    
    //show current and forecast
    this._deck.setAttribute("selectedIndex", "0");
  },
  
  isTodayShown: function()
  {
    var hour = (new Date()).getHours();
    if (this.mBranch.getBoolPref("dayt.panel.enabled")) {
      if (this.mBranch.getIntPref("dayt.panel.mode") == 0 &&
          this.mBranch.getIntPref("dayt.panel.switch") == 2 &&
          hour >= FF_TIME_DAY)
          return false;
      else
          return true;
    }
    
    return false;     
  },
  
  updateAll: function()
  {
    if (this.mUpdate.getValue("lastError","type",0) != "") {
      this.updateCurrent();
      this.updateToday();
      this.updateForecast();
      this.updateLinks();
    }
    else
      this.updateError();
  },

  updateStart: function()
  {
    var wprogress = document.getElementById("ff-progress");

    //set progress meter
    wprogress.setAttribute("mode", "undetermined");
    wprogress.setAttribute("tooltiptext", this._strings.getString("ff.tooltip.loading"));
    wprogress.removeAttribute("value");
    wprogress.removeAttribute("collapsed");

    //show meter
    this._deck.setAttribute("selectedIndex", "1");
  },
  

  updateError: function(type)
  {
    // remove the progress bar (this can stay up if everything gets an error)
    this.removeProgress();
    
    var werror = null;
    
    switch (type) {
      case "CURRENT":
        werror = document.getElementById("ff-current-error");
        werror.setAttribute("value", this.mUpdate.getValue("lastError", "msg", 0));
        werror.setAttribute("tooltiptext", this.mUpdate.getValue("lastError", "tip", 0));
        werror.removeAttribute("collapsed");
        
        // show the error description
        document.getElementById("ff-current-deck").setAttribute("selectedIndex", "1");
        
        // show the spacer if forecast or today is shown
        if (this.mBranch.getBoolPref("dayf.panel.enabled") || this.isTodayShown())
          this._spacer.removeAttribute("collapsed");
        
        break;
        
      case "FORECAST":
        werror = document.getElementById("ff-forecast-error");
        werror.setAttribute("value", this.mUpdate.getValue("lastError", "msg", 0));
        werror.setAttribute("tooltiptext", this.mUpdate.getValue("lastError", "tip", 0));
        werror.removeAttribute("collapsed");
        
        // show the error description
        document.getElementById("ff-forecast-deck").setAttribute("selectedIndex", "1");
          
        break;
        
      case "LINKS":
        var menu  = document.getElementById("ff-links-menu");
        menu.setAttribute("disabled", true);
        break;
    }
    
    // If both current and forecast have the same error, only show forecast
    if ((document.getElementById("ff-current-error").getAttribute("value") == document.getElementById("ff-forecast-error").getAttribute("value")) &&
        (Number(document.getElementById("ff-current-deck").getAttribute("selectedIndex")) == 1)) {
      document.getElementById("ff-current-error").setAttribute("collapsed", true);
      this._current.setAttribute("collapsed", true);
    }
  },
  
  updateCurrent: function()
  {
    this.removeError("current");
    this.removeProgress();
    
    // make sure the current conditions is shown in the deck
    document.getElementById("ff-current-deck").setAttribute("selectedIndex", "0");
      
    //if both forecast and current is shown, uncollapse the spacer
    if ((this.mBranch.getBoolPref("cc.panel.enabled")) && (this.mBranch.getBoolPref("dayf.panel.enabled") || this.isTodayShown()))
      this._spacer.removeAttribute("collapsed");
    else
      this._spacer.setAttribute("collapsed", true);

    //only update if showing
    if (!this.mBranch.getBoolPref("cc.panel.enabled")) {
      this._current.setAttribute("collapsed", "true");
      return
    }

    //if tooltip has been set before remove it
    if (this._current.hasAttribute("tooltip")) {
      var tooltip = document.getElementById(this._current.getAttribute("tooltip"));
      tooltip.parentNode.removeChild(tooltip);
      this._current.removeAttribute("tooltip");
    }
      
    this._current.removeAttribute("collapsed");

    //set generic loc attributes
    count = {};
    keys = this.mUpdate.getKeys("location", count);
    for (y=0; y < count.value; y++)
      this._current.setAttribute(keys[y], this.mUpdate.getValue("location", keys[y], 0));

    //get all attributes for current day
    var count = {};
    var keys = this.mUpdate.getKeys("current", count);
    for (x=0; x < count.value; x++)
      this._current.setAttribute(keys[x], this.mUpdate.getValue("current", keys[x], 0));

    //no info for today
    if (this._current.getAttribute("icon") == "-" || this._current.getAttribute("tmp") == "N/A") {
      this._current.setAttribute("src", "chrome://forecastfox/skin/images/na.png");
      this._current.setAttribute("tooltiptext", this._strings.getString("ff.tooltip.nodata"));
    }
    else {
      this.createTooltip(this._current, "CURRENT");
      this._current.removeAttribute("tooltiptext");        
      switch (this.mBranch.getIntPref("cc.panel.display")) {
        case 0:
        default:
          this._current.setAttribute("src", "chrome://forecastfox/skin/images/small/" + this._current.getAttribute("icon") + ".png");
          this._current.removeAttribute("label");
          break;
        case 1:
          this._current.removeAttribute("src");
          if (this.mBranch.prefHasUserValue("cc.panel.label"))
            this._current.setAttribute("label", this.setLabel(this.getUTFPref("cc.panel.label"), "CURRENT", this._current));
          else
            this._current.setAttribute("label", this.setLabel(this._strings.getString(this.mBranch.getCharPref("cc.panel.label")), "CURRENT", this._current));          
          break;
        case 2:
          this._current.setAttribute("src", "chrome://forecastfox/skin/images/small/" + this._current.getAttribute("icon") + ".png");
          if (this.mBranch.prefHasUserValue("cc.panel.label"))
            this._current.setAttribute("label", this.setLabel(this.getUTFPref("cc.panel.label"), "CURRENT", this._current));
          else
            this._current.setAttribute("label", this.setLabel(this._strings.getString(this.mBranch.getCharPref("cc.panel.label")), "CURRENT", this._current));   
          break;
      }
    }
    this.updateBorders();
  },

  updateToday: function()
  {
    this.removeError("today");
    this.removeProgress();
    
    if (!this.mBranch.getBoolPref("dayt.panel.enabled")) {
      this._today.setAttribute("collapsed", true);
      return;
    }
      
    // make sure the forecast is shown in the deck
    document.getElementById("ff-forecast-deck").setAttribute("selectedIndex", "0");
    
    //remove old today
    while (this._today.hasChildNodes()) {
      if (this._today.lastChild.hasAttribute("tooltip")) {
        tooltip = document.getElementById(this._today.lastChild.getAttribute("tooltip"));
        tooltip.parentNode.removeChild(tooltip);
        this._today.lastChild.removeAttribute("tooltip");
      }    
      this._today.removeChild(this._today.lastChild);
    }
    
    //hide the box and return if we don't show any today forecasts
    var mode    = this.mBranch.getIntPref("dayt.panel.mode");
    
    //if we don't want to show any "today" values, then hide the box and get out
    if (mode == -1) {
      this._today.setAttribute("collapsed", true);
      return;
    }
    
    var sw_mode = this.mBranch.getIntPref("dayt.panel.switch");
    var panels = (mode == 2) ? 2 : 1;
    var cHours = (new Date()).getHours();
    var day_index = 0;
    var inDate    = new Date(this.mUpdate.getValue("forecast", "lsup", day_index));
    
    while (this.shouldIncDay(day_index))
      day_index++;
    
    for (var x = 0; x < panels; x++) {
      
      if (cHours <= FF_TIME_NIGHT || cHours >= FF_TIME_DAY) {
        if (mode == 0 && sw_mode == 1)
          day_index++;
        else if (mode == 2)
          panels--;
      }
      

      if (this.shouldCreateTPanel("day", cHours, x))
        this.createPanel("TODAY", this._today, day_index, "day");
   
      else if (this.shouldCreateTPanel("night", cHours, x))
        this.createPanel("TODAY", this._today, day_index, "night");   
    }
    
    this.updateBorders();    
    
    // show the box!
    this._today.removeAttribute("collapsed");
  },
  
  shouldCreateTPanel: function(part, cHours, x)
  {     
    var mode    = this.mBranch.getIntPref("dayt.panel.mode");
    var sw_mode = this.mBranch.getIntPref("dayt.panel.switch");
        
    switch (part) {
      case "day":
        if (cHours > FF_TIME_NIGHT && cHours < FF_TIME_DAY) {
          if (mode == 0)
            return true;
          else if (mode == 2 && x == 0)
            return true;
          else
            return false;
        }
        else {
          if (mode == 0 && sw_mode == 1)
            return true;
          else if (mode == 2 && x == 2)
            return true;
          else
            return false;
        }
      case "night":
        if (mode == 1)
          return true;
        else if (cHours > FF_TIME_NIGHT && cHours < FF_TIME_DAY) {
          if (mode == 2 && x == 1)
            return true;
          else
            return false;
        }
        else {
          if (mode == 0 && sw_mode == 0)
            return true;
          else if (mode == 2)
            return true;
          else
            return false;
        }
    }
  },

  updateForecast: function()
  {
    // make sure the progress bar and the forecast error are not showing.
    this.removeError("forecast");
    this.removeProgress();
    
    // make sure the forecast is shown in the deck
    document.getElementById("ff-forecast-deck").setAttribute("selectedIndex", "0");
    
    //remove old forecast
    while (this._forecast.hasChildNodes()) {
      if (this._forecast.lastChild.hasAttribute("tooltip")) {
        tooltip = document.getElementById(this._forecast.lastChild.getAttribute("tooltip"));
        tooltip.parentNode.removeChild(tooltip);
        this._forecast.lastChild.removeAttribute("tooltip");
      }    
      this._forecast.removeChild(this._forecast.lastChild);
    }
    
    var days = this.mBranch.getIntPref("dayf.panel.days");
    
    //hide the box and return if we don't show any days
    if (!this.mBranch.getBoolPref("dayf.panel.enabled") || days == 0) {
      this._forecast.setAttribute("collapsed", true);
      return;
    }
    
    // Now fill in the forecast box with the new panels

    var new_panel = false;
    var cHours    = (new Date()).getHours();
    var day_index = 1;
    var day_add   = 0;

    
    while (this.shouldIncDay(day_add))
      day_add++;
      
    day_index += day_add;
        
    for (var x = 0; x < days; day_index++) {

      // finish if we went through too many days
      if (day_index > 9)
        break;

      // create the corresponding nodes
      if (this.shouldCreateFPanel(day_index, "day"))
        new_panel = this.createPanel("FORECAST", this._forecast, day_index, "day");
      
      if (this.shouldCreateFPanel(day_index, "night"))
        new_panel = this.createPanel("FORECAST", this._forecast, day_index, "night");
      
      // increase x if we made a new one
      if (new_panel)
        x++;
      
      new_panel = false;
    }
    
    this.updateBorders();
    
    // uncollapse it!
    this._forecast.removeAttribute("collapsed");
  },
  
  // Removes left border from left-most ff-panel and right border from right-most ff-panel
  updateBorders: function() {
    // first remove old attributes
    var panels = this._deck.getElementsByTagName("statusbarpanel");
    for (var x = 0; x < panels.length; x++) {
      panels[x].removeAttribute("ff-firstChild");
      panels[x].removeAttribute("ff-lastChild");
    }
    
    try {
    
      // now set the ff-firstChild attribute
      if (this.mBranch.getBoolPref("cc.panel.enabled"))
        this._current.setAttribute("ff-firstChild", true);
      else if (this.isTodayShown())
        this._today.firstChild.setAttribute("ff-firstChild", true);
      else if (this.mBranch.getBoolPref("dayf.panel.enabled") &&
              (this.mBranch.getIntPref("dayf.panel.days") != 0))
        this._forecast.firstChild.setAttribute("ff-firstChild", true);
        
      // now set the ff-lastChild attribute
      if (this.mBranch.getBoolPref("dayf.panel.enabled") &&
          (this.mBranch.getIntPref("dayf.panel.days") != 0))
        this._forecast.lastChild.setAttribute("ff-lastChild", true);
      else if (this.isTodayShown())
        this._today.lastChild.setAttribute("ff-lastChild", true);
      else if (this.mBranch.getBoolPref("cc.panel.enabled"))
        this._current.setAttribute("ff-lastChild", true);
    } catch (e) { }
    
  },
  
  // Abstraction for switching days:
  isDataPresent: function(day_index, part)
  {
    // data is not present for that part if there is no temperature
    var temp = (part == "day") ? "hi" : "low";
    return (this.mUpdate.getValue("forecast", temp, day_index) != "N/A");
  },
  
  shouldIncDay: function(day_index)
  {
    var cHours = (new Date()).getHours();
    return (!this.isDataPresent(day_index, "day") && (FF_TIME_NIGHT < cHours && cHours < FF_TIME_DAY));
  },
  
  shouldCreateFPanel: function(day_index, part)
  { 
    var mode      = this.mBranch.getIntPref("dayf.panel.mode");
    var t_mode    = this.mBranch.getIntPref("dayt.panel.mode");
    var t_sw_mode = this.mBranch.getIntPref("dayt.panel.switch");
    var cHours    = (new Date()).getHours();
     
    switch (part) {
      case "night":
        return ((mode == 1) || (mode == 2))
      case "day":
        // if we're displaying days or days and nights
        if (mode == 0 || mode == 2) {
        
          // If its between the "day" hours
          if (FF_TIME_NIGHT < cHours && cHours < FF_TIME_DAY)
            return true;
          // If its between the "night" hours
          else {
            if (t_mode != 0)
              return true;
            else if (day_index != 1 && t_sw_mode == 1)
              return true;
            else if (t_sw_mode != 1)
              return true;
            else
              return false;
          }
        }
        else
          return false;
    }
  },
   
  //////////////////
  
  hideTooltip: function(aNode)
  {
    if (!aNode.hasAttribute("tooltip"))
      return;
      
    var tooltip = document.getElementById(aNode.getAttribute("tooltip"));
    tooltip.hidePopup();
  },
  
  updateLinks: function()
  {
    var popup = document.getElementById("ff-links-popup");
    var menu  = document.getElementById("ff-links-menu");

    //remove old links
    while (popup.hasChildNodes())
      popup.removeChild(popup.lastChild);

    //add links
    var x= this.mUpdate.getLinkCount() -1;
    var pos = 0;
    var newpos, menuitem;
    if (x < 0)
      return;
      
    do {
      menuitem = document.createElement("menuitem");
      menuitem.setAttribute("label", this.mUpdate.getValue("links","title",x));
      menuitem.setAttribute("url", this.mUpdate.getValue("links","url",x));
      menuitem.setAttribute("oncommand","gForecastfox.openPromo(event);");
      menuitem.setAttribute("onclick","checkForMiddleClick(this, event);");
      newpos = Number(this.mUpdate.getValue("links","pos",x));
      menuitem.setAttribute("id", "link" + newpos);  
      if (newpos < pos)
        popup.insertBefore(menuitem, document.getElementById("link" + pos));
      else
        popup.appendChild(menuitem);
      pos = newpos;
    } while (x--);
    
    if (this.mUpdate.getLinkCount() > 0)
      menu.removeAttribute("disabled");
  },

  createTooltip: function(node, mode) 
  {
    if ((mode == "CURRENT" && !this.mBranch.getBoolPref("cc.tooltip.enabled")) ||
        (mode == "FORECAST" && !this.mBranch.getBoolPref("dayf.tooltip.enabled")) ||
        (mode == "TODAY" && !this.mBranch.getBoolPref("dayt.tooltip.enabled")))
      // if the tooltip is not enabled for that type, don't create it.
      return;
     
    var tooltip, parent, lblparent, lbls, lbl, display, image, imgparent;
    
    //create empty tip
    parent = document.getElementById("mainPopupSet");
    tooltip = document.createElement("tooltip");
    tooltip.setAttribute("id", node.getAttribute("id") + "_tip");
    tooltip.setAttribute("noautohide", true);
    
    //set display mode
    if (mode == "CURRENT")
      display = this.mBranch.getIntPref("cc.tooltip.display");
    else if (mode == "FORECAST")
      display = this.mBranch.getIntPref("dayf.tooltip.display");
    else if (mode == "TODAY")
      display = this.mBranch.getIntPref("dayt.tooltip.display");
    
    //set image 
    switch (display) {
      case 0:
        image = document.createElement("image");
        image.setAttribute("src", "chrome://forecastfox/skin/images/large/" + node.getAttribute("icon") + ".png");
        image.setAttribute("class", "ff-tipicon");
        tooltip.appendChild(image);
        break;
      case 1:
        tooltip.setAttribute("orient", "vertical");
        lblparent = tooltip;
        break;
      case 2:
        imgparent = document.createElement("vbox");
        image = document.createElement("image");
        image.setAttribute("src", "chrome://forecastfox/skin/images/large/" + node.getAttribute("icon") + ".png");
        image.setAttribute("class", "ff-tipicon");
        imgparent.appendChild(image);
        lblparent = document.createElement("vbox");
        tooltip.setAttribute("orient", "horizontal");
        tooltip.appendChild(imgparent);
        tooltip.appendChild(lblparent);                             
        break;
    }
    
    //split tooltip
    if (display > 0) {
      if (mode == "CURRENT") { 
        if (this.mBranch.prefHasUserValue("cc.tooltip.label"))
          lbls = this.getUTFPref("cc.tooltip.label");
        else   
          lbls = this._strings.getString(this.mBranch.getCharPref("cc.tooltip.label"));
      }
      else if (mode == "FORECAST") {
        if (this.mBranch.prefHasUserValue("dayf.tooltip.label"))
          lbls = this.getUTFPref("dayf.tooltip.label");
        else   
          lbls = this._strings.getString(this.mBranch.getCharPref("dayf.tooltip.label"));
      } 
      else if (mode == "TODAY") {
        if (this.mBranch.prefHasUserValue("dayt.tooltip.label"))
          lbls = this.getUTFPref("dayt.tooltip.label");
        else   
          lbls = this._strings.getString(this.mBranch.getCharPref("dayt.tooltip.label"));
      }
    
      lbls = lbls.split("/n");  
      for (var i = 0; i < lbls.length; ++i) {
        lbl = document.createElement("label");
        lbl.setAttribute("value", this.setLabel(lbls[i], mode, node));
        lblparent.appendChild(lbl);
      }
    }
    
    //append new tooltip
    try {
      parent.appendChild(tooltip);
    }
    catch (e) {
      parent = document.getElementById("contentAreaContextSet");
      parent.appendChild(tooltip);
    }
    
    node.setAttribute("tooltip", tooltip.getAttribute("id"));
  },
  
  checkNull: function(value)
  {
    return (value == null) ? "N/A" : value;
  },
  
  createPanel: function(type, parent, day_index, part)
  {
    var node = document.createElement("statusbarpanel");
    var prefix = (type == "FORECAST") ? "dayf" : "dayt";
    //set context menu on node
    node.setAttribute("context","ff-popup");
      
    //creates a forecast panel
    var keys, count;
    
    //set generic loc attributes
    count = {};
    keys = this.mUpdate.getKeys("location", count);
    for (y=0; y < count.value; y++)
      node.setAttribute(keys[y], this.checkNull(this.mUpdate.getValue("location", keys[y], day_index)));
    //set generic forecast attributes
    count = {};
    keys = this.mUpdate.getKeys("forecast", count);
    for (y=0; y < count.value; y++)
      node.setAttribute(keys[y], this.checkNull(this.mUpdate.getValue("forecast", keys[y], day_index)));

    // set night/day specific attributes
    switch (part) {
      case "night":
        //set night attributes
        count = {};
        keys = this.mUpdate.getKeys("night", count);
        for (y=0; y < count.value; y++)
          node.setAttribute(keys[y], this.checkNull(this.mUpdate.getValue("night", keys[y], day_index)));
        break;
        
      case "day":
        //set day attributes
        count = {};
        keys = this.mUpdate.getKeys("day", count);
        for (y=0; y < count.value; y++)
          node.setAttribute(keys[y], this.checkNull(this.mUpdate.getValue("day", keys[y], day_index)));
        break;
    }
      
    // set the image, class and tooltip
    node.setAttribute("part", part);
    node.setAttribute("id", "forecast" + part + day_index);
    node.setAttribute("dayNum", day_index);
    node.setAttribute("class", "ff-panel");
    node.setAttribute("onmouseout", "gForecastfox.hideTooltip(this);"); 
    this.createTooltip(node, type);
    
    switch (this.mBranch.getIntPref(prefix + ".panel.display")) {
      case 0:
      default:
        node.setAttribute("src", "chrome://forecastfox/skin/images/small/" + node.getAttribute("icon") + ".png");
        node.removeAttribute("label");
        break;
      case 1:
        node.removeAttribute("src");
        if (this.mBranch.prefHasUserValue(prefix + ".panel.label"))
          node.setAttribute("label", this.setLabel(this.getUTFPref(prefix + ".panel.label"), "FORECAST", node));
        else
          node.setAttribute("label", this.setLabel(this._strings.getString(this.mBranch.getCharPref(prefix + ".panel.label")), "FORECAST", node)); 
        break;
      case 2:
        node.setAttribute("src", "chrome://forecastfox/skin/images/small/" + node.getAttribute("icon") + ".png");
        if (this.mBranch.prefHasUserValue(prefix + ".panel.label"))
          node.setAttribute("label", this.setLabel(this.getUTFPref(prefix + ".panel.label"), "FORECAST", node));
        else
          node.setAttribute("label", this.setLabel(this._strings.getString(this.mBranch.getCharPref(prefix + ".panel.label")), "FORECAST", node)); 
        break;
    }

    parent.appendChild(node);
    
    // return that one was made
    return true;
  },
  
  populateProfiles: function()
  {
    var prof_menu = document.getElementById("ff-profs-popup");

    //remove old profiles
    while (prof_menu.hasChildNodes())
      prof_menu.removeChild(prof_menu.lastChild);
    
    var count = {};
    var keys = this.mProfile.getNames(count);
    var item;

    if (count.value == 0) {
      item = document.createElement("menuitem");
      item.setAttribute("label", this._strings.getString("ff.profiles.none"));
      item.setAttribute("disabled", true);
      prof_menu.appendChild(item);
      return;
    }
    
    for (x = 0; x < count.value; x++) {
      item = document.createElement("menuitem");
      item.setAttribute("label", keys[x]);
      item.setAttribute("oncommand", "gForecastfox.switchProfile(this)"); 
      item.setAttribute("type", "radio");
      item.setAttribute("name", "profile");     
      if (keys[x] == this.mBranch.getCharPref("profile.current"))
        item.setAttribute("checked", "true");
      prof_menu.appendChild(item);
    }

    return;
  },
  
  switchProfile: function(item)
  {
    document.getElementById("ff-profs-popup").hidePopup();
    var new_profile = item.getAttribute("label");
    if (new_profile != this.mBranch.getCharPref("profile.current"))
      this.mBranch.setCharPref("profile.current", new_profile);
  },
  
  mainPopupShowing: function(aEvent)
  {
    // save the event for if the forecast button is clicked...
    this.mLastEvent = {};
    this.mLastEvent.explicitOriginalTarget = aEvent.explicitOriginalTarget;
    
    if (aEvent.explicitOriginalTarget.getAttribute("id") == "ff-current-panel")
      this.mLastEvent.currentTarget = aEvent.explicitOriginalTarget;
    else
      this.mLastEvent.currentTarget = aEvent.explicitOriginalTarget.parentNode;
    
    // if its the context menu for the error message or progress menu don't show the forecast item
    var target = aEvent.explicitOriginalTarget.getAttribute("id");
    
    if (target.match("error") || target == "ff-progress")
      document.getElementById("ff-popup-forecast").setAttribute("collapsed", true);
    else
      document.getElementById("ff-popup-forecast").removeAttribute("collapsed");
  },
  
  getLastEvent: function()
  {
    return this.mLastEvent;
  },
  
  openPromo: function(aEvent)
  {
    var url = aEvent.currentTarget.getAttribute("url");
    try {
      openUILink(url, aEvent);
    } catch (e) {
      window._content.document.location = url;
    }
  },

  openLink: function(aEvent, aLoc, aOpenEvent)
  {
    // supply an aOpenEvent and it will be used in the openUILink instead of the aEvent

    var url = aEvent.currentTarget.getAttribute("url");
    if (aLoc) {
      var locid = this.mBranch.getCharPref("general.locid");
      url = url + locid;
    }

    if (url.indexOf("?") == -1)
      url = url + "?";
    else
      url = url + "&";
    
    if ((aEvent.currentTarget.getAttribute("id") != "ff-current-panel") &&
        (aEvent.currentTarget.getAttribute("id") != "ff-popup-twc"))
      url = url + "dayNum=" + aEvent.explicitOriginalTarget.getAttribute("dayNum") + "&";
          
    url = url + "prod=" + this.mUpdate.productCode + "&par=" + this.mUpdate.partnerId;
    
    // if aOpenEvent exists, use it instead of aEvent
    if (aOpenEvent)
      aEvent = aOpenEvent;
    
    try {
      openUILink(url, aEvent);
    }
    catch (e) {
      window._content.document.location = url;
    }
  },
  
  setLabel: function(text, type, element) 
  {
    if (!text)
      return "";

    var tool_base = [[/\[locale\]/g, "element.getAttribute('locale')"],
                     [/\[prof\]/g, "this.mBranch.getCharPref('profile.current')"],
                     [/\[ut\]/g, "element.getAttribute('ut')"],
                     [/\[ud\]/g, "element.getAttribute('ud')"],
                     [/\[us\]/g, "element.getAttribute('us')"],
                     [/\[up\]/g, "element.getAttribute('up')"],
                     [/\[ur\]/g, "element.getAttribute('ur')"],
                     [/\[dnam\]/g, "element.getAttribute('dnam')"],
                     [/\[tm\]/g, "element.getAttribute('tm')"],
                     [/\[lat\]/g, "element.getAttribute('lat')"],
                     [/\[lon\]/g, "element.getAttribute('lon')"],
                     [/\[sunr\]/g, "element.getAttribute('sunr')"],
                     [/\[suns\]/g, "element.getAttribute('suns')"],
                     [/\[zone\]/g, "element.getAttribute('zone')"],
                     [/\[deg\]/g, "String.fromCharCode(176)"]];
    var tool_append = null;

    if (type == "CURRENT") {
      tool_append = [[/\[lsup\]/g, "element.getAttribute('lsup')"],
                    [/\[obst\]/g, "element.getAttribute('obst')"],
                    [/\[tmp\]/g, "element.getAttribute('tmp')"],
                    [/\[flik\]/g, "element.getAttribute('flik')"],
                    [/\[t\]/g, "element.getAttribute('t')"],
                    [/\[icon\]/g, "element.getAttribute('icon')"],
                    [/\[barr\]/g, "element.getAttribute('barr')"],
                    [/\[bard\]/g, "element.getAttribute('bard')"],
                    [/\[winds\]/g, "element.getAttribute('winds')"],
                    [/\[windgust\]/g, "element.getAttribute('windgust')"],
                    [/\[windd\]/g, "element.getAttribute('windd')"],
                    [/\[windt\]/g, "element.getAttribute('windt')"],
                    [/\[hmid\]/g, "element.getAttribute('hmid')"],
                    [/\[vis\]/g, "element.getAttribute('vis')"],
                    [/\[uvi\]/g, "element.getAttribute('uvi')"],
                    [/\[uvt\]/g, "element.getAttribute('uvt')"],
                    [/\[moont\]/g, "element.getAttribute('moont')"],
                    [/\[dewp\]/g, "element.getAttribute('dewp')"]];
    }
    else if (type == "FORECAST" || type == "TODAY") {
      tool_append = [[/\[part\]/g, "element.getAttribute('part')"],
                    [/\[day\]/g, "element.getAttribute('day')"],
                    [/\[date\]/g, "element.getAttribute('date')"],
                    [/\[lsup\]/g, "element.getAttribute('lsup')"],
                    [/\[hi\]/g, "element.getAttribute('hi')"],
                    [/\[low\]/g, "element.getAttribute('low')"],
                    [/\[sunr\]/g, "element.getAttribute('sunr')"],
                    [/\[suns\]/g, "element.getAttribute('suns')"],
                    [/\[icon\]/g, "element.getAttribute('icon')"],
                    [/\[t\]/g, "element.getAttribute('t')"],
                    [/\[winds\]/g, "element.getAttribute('winds')"],
                    [/\[windgust\]/g, "element.getAttribute('windgust')"],
                    [/\[windd\]/g, "element.getAttribute('windd')"],
                    [/\[windt\]/g, "element.getAttribute('windt')"],
                    [/\[ppcp\]/g, "element.getAttribute('ppcp')"],
                    [/\[hmid\]/g, "element.getAttribute('hmid')"],
                    [/\[sday\]/g, "element.getAttribute('day').substr(0,3)" ],
                    [/\[tmp\]/g, "(element.getAttribute('part') == 'day') ? element.getAttribute('hi') : element.getAttribute('low')"]];
    }
    var tool_array = concat(tool_base, tool_append);
    
    // Then replace each one with the value.
    for (var x = 0; x < tool_array.length; x++)
        text = text.replace(tool_array[x][0], eval(tool_array[x][1]));

    return text;
  },
  
  checkForMiddleClick: function (node, event)
  {
    if (event.button == 1) {
      eval(node.getAttribute("ondblclick"));
      closeMenus(event.target);
    }
  }
}

function concat(c1, c2)
{
  // Concats too collections into an array.
  var c3 = new Array(c1.length + c2.length);
  var x,y = 0;

  for (x = 0; x < c1.length; x++)
    c3[y++] = c1[x];

  for (x = 0; x < c2.length; x++)
    c3[y++] = c2[x];

  return c3;
}
